@@ -41,6 +41,7 @@
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
 #![allow(clippy::needless_lifetimes)] // See arti#1765
+#![allow(mismatched_lifetime_syntaxes)] // temporary workaround for arti#2060
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 // TODO #1645 (either remove this, or decide to have it everywhere)
@@ -116,7 +117,11 @@ use clap::{value_parser, Arg, ArgAction, Command};
 #[allow(unused_imports)]
 use tracing::{error, info, warn};
 
-#[cfg(any(feature = "hsc", feature = "onion-service-service"))]
+#[cfg(any(
+    feature = "hsc",
+    feature = "onion-service-service",
+    feature = "onion-service-cli-extra"
+))]
 use clap::Subcommand as _;
 
 #[cfg(feature = "experimental-api")]
@@ -294,6 +299,12 @@ where
         }
     }
 
+    cfg_if::cfg_if! {
+        if #[cfg(feature = "onion-service-cli-extra")] {
+            let clap_app = subcommands::keys::KeysSubcommands::augment_subcommands(clap_app);
+        }
+    }
+
     // Tracing doesn't log anything when there is no subscriber set.  But we want to see
     // logging messages from config parsing etc.  We can't set the global default subscriber
     // because we can only set it once.  The other ways involve a closure.  So we have a
@@ -384,6 +395,15 @@ where
         return subcommands::proxy::run(runtime, proxy_matches, cfg_sources, config, client_config);
     }
 
+    // Check for the optional "keys" subcommand.
+    cfg_if::cfg_if! {
+        if #[cfg(feature = "onion-service-cli-extra")] {
+            if let Some(keys_matches) = matches.subcommand_matches("keys") {
+                return subcommands::keys::run(runtime, keys_matches, &client_config);
+            }
+        }
+    }
+
     // Check for the optional "hss" subcommand.
     cfg_if::cfg_if! {
         if #[cfg(feature = "onion-service-service")] {
