@@ -6,7 +6,7 @@ pub(crate) mod hss;
 #[cfg(feature = "hsc")]
 pub(crate) mod hsc;
 
-#[cfg(feature = "relay")]
-pub(crate) mod relay;
+#[cfg(feature = "onion-service-cli-extra")]
+pub(crate) mod keys;
 
 pub(crate) mod proxy;
