@@ -101,9 +101,10 @@ pub(crate) struct CommonArgs {
     )]
     key_type: KeyType,
 
-    /// Whether to write the prompt or not
+    /// With this flag active no prompt will be shown
+    /// and no confirmation will be asked
     #[arg(long, short, default_value_t = false)]
-    quiet: bool,
+    batch: bool,
 }
 
 /// The common arguments of the key subcommands.
@@ -128,10 +129,6 @@ pub(crate) struct RotateKeyArgs {
     /// Arguments for configuring keygen.
     #[command(flatten)]
     keygen: KeygenArgs,
-
-    /// Do not prompt before overwriting the key.
-    #[arg(long, short)]
-    force: bool,
 }
 
 /// The arguments of the [`Remove`](KeySubcommand::Remove) subcommand.
@@ -140,10 +137,6 @@ pub(crate) struct RemoveKeyArgs {
     /// Arguments shared by all hsc subcommands.
     #[command(flatten)]
     common: CommonArgs,
-
-    /// Do not prompt before removing the key.
-    #[arg(long, short)]
-    force: bool,
 }
 
 /// Run the `hsc` subcommand.
@@ -184,7 +177,7 @@ fn run_key(subcommand: KeySubcommand, client: &InertTorClient) -> Result<()> {
 
 /// Run the `hsc prepare-stealth-mode-key` subcommand.
 fn prepare_service_discovery_key(args: &GetKeyArgs, client: &InertTorClient) -> Result<()> {
-    let addr = get_onion_address(args.common.quiet)?;
+    let addr = get_onion_address(&args.common)?;
     let key = match args.generate {
         GenerateKey::IfNeeded => {
             // TODO: consider using get_or_generate in generate_service_discovery_key
@@ -251,12 +244,10 @@ fn write_public_key(mut f: impl io::Write, key: &HsClientDescEncKey) -> io::Resu
 
 /// Run the `hsc rotate-key` subcommand.
 fn rotate_service_discovery_key(args: &RotateKeyArgs, client: &InertTorClient) -> Result<()> {
-    let addr = get_onion_address(args.common.quiet)?;
-    if !args.force {
-        let msg = format!("rotate client restricted discovery key for {}?", addr);
-        if !prompt(&msg)? {
-            return Ok(());
-        }
+    let addr = get_onion_address(&args.common)?;
+    let msg = format!("rotate client restricted discovery key for {}?", addr);
+    if !prompt(&msg, &args.common)? {
+        return Ok(());
     }
 
     let key = client.rotate_service_discovery_key(KeystoreSelector::default(), addr)?;
@@ -266,12 +257,10 @@ fn rotate_service_discovery_key(args: &RotateKeyArgs, client: &InertTorClient) -
 
 /// Run the `hsc remove-key` subcommand.
 fn remove_service_discovery_key(args: &RemoveKeyArgs, client: &InertTorClient) -> Result<()> {
-    let addr = get_onion_address(args.common.quiet)?;
-    if !args.force {
-        let msg = format!("remove client restricted discovery key for {}?", addr);
-        if !prompt(&msg)? {
-            return Ok(());
-        }
+    let addr = get_onion_address(&args.common)?;
+    let msg = format!("remove client restricted discovery key for {}?", addr);
+    if !prompt(&msg, &args.common)? {
+        return Ok(());
     }
 
     let _key = client.remove_service_discovery_key(KeystoreSelector::default(), addr)?;
@@ -283,26 +272,35 @@ fn remove_service_discovery_key(args: &RemoveKeyArgs, client: &InertTorClient) -
 ///
 /// Loops until the user confirms or declines,
 /// returning true if they confirmed.
-fn prompt(msg: &str) -> Result<bool> {
+///
+/// If `args.batch` is `true` no confirmation will be asked.
+fn prompt(msg: &str, args: &CommonArgs) -> Result<bool> {
+    if args.batch {
+        return Ok(true);
+    }
+
     /// The accept message.
     const YES: &str = "YES";
     /// The decline message.
     const NO: &str = "no";
 
-    let msg = format!("{msg} (type {YES} or {NO})");
-    loop {
-        let proceed = dialoguer::Input::<String>::new()
-            .with_prompt(&msg)
-            .interact_text()?;
+    let mut proceed = String::new();
 
-        let proceed: &str = proceed.as_ref();
-        if proceed == YES {
+    print!("{} (type {YES} or {NO}): ", msg);
+    io::stdout().flush().map_err(|e| anyhow!(e))?;
+    loop {
+        io::stdin()
+            .read_line(&mut proceed)
+            .map_err(|e| anyhow!(e))?;
+
+        if proceed.trim_end() == YES {
             return Ok(true);
         }
 
-        match proceed.to_lowercase().as_str() {
+        match proceed.trim_end().to_lowercase().as_str() {
             NO | "n" => return Ok(false),
             _ => {
+                proceed.clear();
                 continue;
             }
         }
@@ -310,9 +308,9 @@ fn prompt(msg: &str) -> Result<bool> {
 }
 
 /// Prompt the user for an onion address.
-fn get_onion_address(quiet: bool) -> Result<HsId, anyhow::Error> {
+fn get_onion_address(args: &CommonArgs) -> Result<HsId, anyhow::Error> {
     let mut addr = String::new();
-    if !quiet {
+    if !args.batch {
         print!("Enter an onion address: ");
         io::stdout().flush().map_err(|e| anyhow!(e))?;
     };
