@@ -34,11 +34,14 @@
 #![deny(clippy::unnecessary_wraps)]
 #![warn(clippy::unseparated_literal_suffix)]
 #![deny(clippy::unwrap_used)]
+#![deny(clippy::mod_module_files)]
 #![allow(clippy::let_unit_value)] // This can reasonably be done for explicitness
 #![allow(clippy::uninlined_format_args)]
 #![allow(clippy::significant_drop_in_scrutinee)] // arti/-/merge_requests/588/#note_2812945
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
+#![allow(clippy::needless_lifetimes)] // See arti#1765
+#![allow(mismatched_lifetime_syntaxes)] // temporary workaround for arti#2060
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 // TODO: Try making it not Deref and having expose+expose_mut instead; how bad is it?
@@ -67,9 +70,8 @@ pub type Result<T> = std::result::Result<T, Error>;
 ///
 /// This behavior can be overridden locally by using
 /// [`with_safe_logging_suppressed`] and globally with [`disable_safe_logging`].
-#[derive(Educe)]
+#[derive(Educe, Clone, Copy)]
 #[educe(
-    Clone(bound),
     Default(bound),
     Deref,
     DerefMut,
@@ -244,9 +246,8 @@ impl<'a, T: Redactable + ?Sized> Redactable for &'a T {
 }
 
 /// A wrapper around a `Redactable` that displays it in redacted format.
-#[derive(Educe)]
+#[derive(Educe, Clone, Copy)]
 #[educe(
-    Clone(bound),
     Default(bound),
     Deref,
     DerefMut,
