@@ -24,7 +24,7 @@ use tor_config::ReconfigureError;
 use tor_error::{error_report, internal, into_internal};
 use tor_netdir::{DirEvent, NetDir, NetDirProvider, Timeliness};
 use tor_persist::{DynStorageHandle, StateMgr};
-use tor_relay_selection::RelayExclusion;
+use tor_relay_selection::RelaySelector;
 use tor_rtcompat::Runtime;
 use tracing::{debug, info};
 
@@ -224,10 +224,10 @@ impl<R: Runtime> VanguardMgr<R> {
 
     /// Return a [`Vanguard`] relay for use in the specified layer.
     ///
-    /// The `neighbor_exclusion` must contain the relays that would neighbor this vanguard
+    /// The `relay_selector` must exclude the relays that would neighbor this vanguard
     /// in the path.
     ///
-    /// Specifically, it should contain
+    /// Specifically, it should exclude
     ///   * the last relay in the path (the one immediately preceding the vanguard): the same relay
     ///     cannot be used in consecutive positions in the path (a relay won't let you extend the
     ///     circuit to itself).
@@ -265,7 +265,7 @@ impl<R: Runtime> VanguardMgr<R> {
         rng: &mut Rng,
         netdir: &'a NetDir,
         layer: Layer,
-        neighbor_exclusion: &RelayExclusion<'a>,
+        relay_selector: &RelaySelector<'a>,
     ) -> Result<Vanguard<'a>, VanguardMgrError> {
         use VanguardMode::*;
 
@@ -285,12 +285,12 @@ impl<R: Runtime> VanguardMgr<R> {
                 (Layer::Layer2, Full) | (Layer::Layer2, Lite) => inner
                     .vanguard_sets
                     .l2()
-                    .pick_relay(rng, netdir, neighbor_exclusion),
+                    .pick_relay(rng, netdir, relay_selector),
                 (Layer::Layer3, Full) => {
                     inner
                         .vanguard_sets
                         .l3()
-                        .pick_relay(rng, netdir, neighbor_exclusion)
+                        .pick_relay(rng, netdir, relay_selector)
                 }
                 _ => {
                     return Err(VanguardMgrError::LayerNotSupported {
@@ -582,10 +582,10 @@ impl VanguardMgr<MockRuntime> {
 #[non_exhaustive]
 pub enum Layer {
     /// L2 vanguard.
-    #[display(fmt = "layer 2")]
+    #[display("layer 2")]
     Layer2,
     /// L3 vanguard.
-    #[display(fmt = "layer 3")]
+    #[display("layer 3")]
     Layer3,
 }
 
@@ -609,6 +609,7 @@ mod test {
 
     use set::TimeBoundVanguard;
     use tor_config::ExplicitOrAuto;
+    use tor_relay_selection::RelayExclusion;
 
     use super::*;
 
@@ -669,6 +670,14 @@ mod test {
         }
     }
 
+    /// Return a maximally permissive RelaySelector for a vanguard.
+    fn permissive_selector() -> RelaySelector<'static> {
+        RelaySelector::new(
+            tor_relay_selection::RelayUsage::vanguard(),
+            RelayExclusion::no_relays_excluded(),
+        )
+    }
+
     /// Look up the vanguard in the specified VanguardSet.
     fn find_in_set<R: Runtime>(
         relay_ids: &RelayIds,
@@ -781,14 +790,13 @@ mod test {
             let vanguardmgr = VanguardMgr::new_testing(&rt, VanguardMode::Lite).unwrap();
             let netdir = testnet::construct_netdir().unwrap_if_sufficient().unwrap();
             let mut rng = testing_rng();
-            let exclusion = RelayExclusion::no_relays_excluded();
             // Wait until the vanguard manager has bootstrapped
             // (otherwise we'll get a BootstrapRequired error)
             let _netdir_provider = vanguardmgr.init_vanguard_sets(&netdir).await.unwrap();
 
             // Cannot select an L3 vanguard when running in "Lite" mode.
             let err = vanguardmgr
-                .select_vanguard(&mut rng, &netdir, Layer3, &exclusion)
+                .select_vanguard(&mut rng, &netdir, Layer3, &permissive_selector())
                 .unwrap_err();
             assert!(
                 matches!(
@@ -809,7 +817,6 @@ mod test {
             let vanguardmgr = VanguardMgr::new_testing(&rt, VanguardMode::Lite).unwrap();
             let netdir = testnet::construct_netdir().unwrap_if_sufficient().unwrap();
             let mut rng = testing_rng();
-            let exclusion = RelayExclusion::no_relays_excluded();
 
             // The sets are initially empty
             assert_sets_empty(&vanguardmgr);
@@ -817,7 +824,7 @@ mod test {
             // VanguardMgr::launch_background tasks was not called, so select_vanguard will return
             // an error (because the vanguard sets are empty)
             let err = vanguardmgr
-                .select_vanguard(&mut rng, &netdir, Layer2, &exclusion)
+                .select_vanguard(&mut rng, &netdir, Layer2, &permissive_selector())
                 .unwrap_err();
 
             assert!(
@@ -840,7 +847,6 @@ mod test {
             let netdir = testnet::construct_netdir().unwrap_if_sufficient().unwrap();
             let params = VanguardParams::try_from(netdir.params()).unwrap();
             let mut rng = testing_rng();
-            let exclusion = RelayExclusion::no_relays_excluded();
 
             // The sets are initially empty
             assert_sets_empty(&vanguardmgr);
@@ -851,7 +857,7 @@ mod test {
             assert_sets_filled(&vanguardmgr, &params);
 
             let vanguard1 = vanguardmgr
-                .select_vanguard(&mut rng, &netdir, Layer2, &exclusion)
+                .select_vanguard(&mut rng, &netdir, Layer2, &permissive_selector())
                 .unwrap();
             assert_expiry_in_bounds(&vanguard1, &vanguardmgr, &rt, &params, Layer2);
 
@@ -862,9 +868,11 @@ mod test {
                     .map(|id| id.to_owned())
                     .collect(),
             );
+            let selector =
+                RelaySelector::new(tor_relay_selection::RelayUsage::vanguard(), exclusion);
 
             let vanguard2 = vanguardmgr
-                .select_vanguard(&mut rng, &netdir, Layer3, &exclusion)
+                .select_vanguard(&mut rng, &netdir, Layer3, &selector)
                 .unwrap();
 
             assert_expiry_in_bounds(&vanguard2, &vanguardmgr, &rt, &params, Layer3);
@@ -884,7 +892,7 @@ mod test {
         netdir_provider: &TestNetDirProvider,
         params: impl IntoIterator<Item = (&str, i32)>,
     ) -> VanguardParams {
-        let new_netdir = testnet::construct_custom_netdir_with_params(|_, _| {}, params, None)
+        let new_netdir = testnet::construct_custom_netdir_with_params(|_, _, _| {}, params, None)
             .unwrap()
             .unwrap_if_sufficient()
             .unwrap();
@@ -940,7 +948,7 @@ mod test {
         netdir_provider: &TestNetDirProvider,
     ) -> NetDir {
         let new_netdir = construct_custom_netdir_with_params(
-            |_idx, bld| {
+            |_idx, bld, _| {
                 let md_so_far = bld.md.testing_md().unwrap();
                 if md_so_far.ed25519_id() == vanguard.relay().id() {
                     bld.omit_rs = true;
@@ -1099,7 +1107,7 @@ mod test {
             let vanguardmgr = VanguardMgr::new_testing(&rt, VanguardMode::Lite).unwrap();
 
             let netdir =
-                construct_custom_netdir_with_params(|_, _| {}, ENABLE_LITE_VANGUARDS, None)
+                construct_custom_netdir_with_params(|_, _, _| {}, ENABLE_LITE_VANGUARDS, None)
                     .unwrap()
                     .unwrap_if_sufficient()
                     .unwrap();
@@ -1111,9 +1119,8 @@ mod test {
             assert!(vanguardmgr.storage.load().unwrap().is_none());
 
             let mut rng = testing_rng();
-            let exclusion = RelayExclusion::no_relays_excluded();
             assert!(vanguardmgr
-                .select_vanguard(&mut rng, &netdir, Layer3, &exclusion)
+                .select_vanguard(&mut rng, &netdir, Layer3, &permissive_selector())
                 .is_err());
 
             // Enable full vanguards again.
@@ -1124,7 +1131,7 @@ mod test {
 
             let vanguard_sets_orig = vanguardmgr.storage.load().unwrap();
             assert!(vanguardmgr
-                .select_vanguard(&mut rng, &netdir, Layer3, &exclusion)
+                .select_vanguard(&mut rng, &netdir, Layer3, &permissive_selector())
                 .is_ok());
 
             // Switch to lite vanguards.
@@ -1144,9 +1151,8 @@ mod test {
             switch_hs_mode_config(&vanguardmgr, VanguardMode::Lite);
 
             let mut rng = testing_rng();
-            let exclusion = RelayExclusion::no_relays_excluded();
             let excluded_vanguard = vanguardmgr
-                .select_vanguard(&mut rng, &netdir, Layer2, &exclusion)
+                .select_vanguard(&mut rng, &netdir, Layer2, &permissive_selector())
                 .unwrap();
 
             let _ = install_netdir_excluding_vanguard(
