@@ -27,7 +27,7 @@ use tor_rtcompat::Runtime;
 use crate::isol_map;
 use crate::{ConnError, HsClientConnector, HsClientSecretKeys};
 
-slotmap::new_key_type! {
+slotmap_careful::new_key_type! {
     struct TableIndex;
 }
 
@@ -136,7 +136,7 @@ const RETAIN_DATA_AFTER_LAST_USE: Duration = Duration::from_secs(48 * 3600 /*hou
 /// In other words,
 ///  * Two HS connection requests cannot share state and effort
 ///    (descriptor downloads, descriptors, intro pt history)
-///    unless the client authg keys to be used are the same.
+///    unless the restricted discovery keys to be used are the same.
 ///  * This criterion is checked before looking at isolations,
 ///    which may further restrict sharing:
 ///    Two HS connection requests will only share state subject to isolations.
@@ -606,8 +606,11 @@ impl<D: MockableConnectorData> Services<D> {
                     now <= expiry_time
                 }
                 ServiceState::Open { .. } | ServiceState::Working { .. } => true,
-                ServiceState::Dummy { .. } => {
-                    error!("found dummy data during HS housekeeping, for {}", sv(hsid));
+                ServiceState::Dummy => {
+                    error!(
+                        "bug: found dummy data during HS housekeeping, for {}",
+                        sv(hsid)
+                    );
                     false
                 }
             });
@@ -634,7 +637,7 @@ impl<D: MockableConnectorData> ServiceState<D> {
             let expiry = last_used
                 .checked_add(RETAIN_CIRCUIT_AFTER_LAST_USE)
                 .or_else(|| {
-                    error!("time overflow calculating HS circuit expiry, killing circuit!");
+                    error!("bug: time overflow calculating HS circuit expiry, killing circuit!");
                     None
                 })?;
             let wait = expiry.checked_duration_since(now).unwrap_or_default();
@@ -748,6 +751,8 @@ pub(crate) mod test {
     use std::task::Poll::{self, *};
     use tokio::pin;
     use tokio_crate as tokio;
+    use tor_memquota::ArcMemoryQuotaTrackerExt as _;
+    use tor_proto::memquota::ToplevelAccount;
     use tor_rtcompat::{test_with_one_runtime, SleepProvider};
     use tor_rtmock::MockRuntime;
     use tracing_test::traced_test;
@@ -858,6 +863,7 @@ pub(crate) mod test {
             &Default::default(),
             tor_chanmgr::Dormancy::Dormant,
             &Default::default(),
+            ToplevelAccount::new_noop(),
         );
         let guardmgr = tor_guardmgr::GuardMgr::new(
             runtime.clone(),
@@ -872,7 +878,7 @@ pub(crate) mod test {
                 tor_persist::TestingStateMgr::new(),
                 &runtime,
                 Arc::new(chanmgr),
-                guardmgr,
+                &guardmgr,
             )
             .unwrap(),
         );
