@@ -34,11 +34,14 @@
 #![deny(clippy::unnecessary_wraps)]
 #![warn(clippy::unseparated_literal_suffix)]
 #![deny(clippy::unwrap_used)]
+#![deny(clippy::mod_module_files)]
 #![allow(clippy::let_unit_value)] // This can reasonably be done for explicitness
 #![allow(clippy::uninlined_format_args)]
 #![allow(clippy::significant_drop_in_scrutinee)] // arti/-/merge_requests/588/#note_2812945
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
+#![allow(clippy::needless_lifetimes)] // See arti#1765
+#![allow(mismatched_lifetime_syntaxes)] // temporary workaround for arti#2060
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 #![allow(dead_code, unused_variables)]
 
@@ -118,8 +121,8 @@ define_bytes! {
 pub struct RendCookie([u8; REND_COOKIE_LEN]);
 }
 
-impl rand::distributions::Distribution<RendCookie> for rand::distributions::Standard {
+impl rand::distr::Distribution<RendCookie> for rand::distr::StandardUniform {
     fn sample<R: rand::Rng + ?Sized>(&self, rng: &mut R) -> RendCookie {
-        RendCookie(rng.gen::<[u8; REND_COOKIE_LEN]>().into())
+        RendCookie(rng.random::<[u8; REND_COOKIE_LEN]>().into())
     }
 }
