@@ -14,6 +14,7 @@ use tor_cell::relaycell::{
     hs::est_intro::{self, EstablishIntroDetails},
     msg::IntroEstablished,
 };
+use tor_proto::TargetHop;
 
 /// Handle onto the task which is establishing and maintaining one IPT
 pub(crate) struct IptEstablisher {
@@ -732,9 +733,6 @@ impl<R: Runtime> Reactor<R> {
             // longer than necessary.
             (protovers, circuit, ipt_details)
         };
-        let intro_pt_hop = circuit
-            .last_hop_num()
-            .map_err(into_internal!("Somehow built a circuit with no hops!?"))?;
 
         let establish_intro = {
             let ipt_sid_id = (*self.k_sid).as_ref().verifying_key().into();
@@ -748,7 +746,8 @@ impl<R: Runtime> Reactor<R> {
                 }
             }
             let circuit_binding_key = circuit
-                .binding_key(intro_pt_hop)
+                .binding_key(TargetHop::LastHop)
+                .await
                 .map_err(IptEstablisherError::CircuitState)?
                 .ok_or(internal!("No binding key for introduction point!?"))?;
             let body: Vec<u8> = details
@@ -792,7 +791,7 @@ impl<R: Runtime> Reactor<R> {
             replay_log,
         };
         let _conversation = circuit
-            .start_conversation(Some(establish_intro), handler, intro_pt_hop)
+            .start_conversation(Some(establish_intro), handler, TargetHop::LastHop)
             .await
             .map_err(IptEstablisherError::SendEstablishIntro)?;
         // At this point, we have `await`ed for the Conversation to exist, so we know
