@@ -182,6 +182,9 @@ pub(super) struct Reactor<R: Runtime, M: Mockable> {
     shutdown_tx: broadcast::Sender<Void>,
     /// Path resolver for configuration files.
     path_resolver: Arc<CfgPathResolver>,
+    /// Queue on which we receive messages from the [`PowManager`] telling us that a seed has
+    /// rotated and thus we need to republish the descriptor for a particular time period.
+    update_from_pow_manager_rx: mpsc::Receiver<TimePeriod>,
 }
 
 /// The immutable, shared state of the descriptor publisher reactor.
@@ -199,6 +202,8 @@ struct Immutable<R: Runtime, M: Mockable> {
     keymgr: Arc<KeyMgr>,
     /// A sender for updating the status of the onion service.
     status_tx: PublisherStatusSender,
+    /// Proof-of-work state.
+    pow_manager: Arc<PowManager<R>>,
 }
 
 impl<R: Runtime, M: Mockable> Immutable<R, M> {
@@ -336,7 +341,7 @@ impl<R: Runtime> Mockable for Real<R> {
     type ClientCirc = ClientCirc;
 
     fn thread_rng(&self) -> Self::Rng {
-        rand::thread_rng()
+        rand::rng()
     }
 
     async fn get_or_launch_specific<T>(
@@ -551,6 +556,8 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
         status_tx: PublisherStatusSender,
         keymgr: Arc<KeyMgr>,
         path_resolver: Arc<CfgPathResolver>,
+        pow_manager: Arc<PowManager<R>>,
+        update_from_pow_manager_rx: mpsc::Receiver<TimePeriod>,
     ) -> Self {
         /// The maximum size of the upload completion notifier channel.
         ///
@@ -581,6 +588,7 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
             nickname,
             keymgr,
             status_tx,
+            pow_manager,
         };
 
         let inner = Inner {
@@ -607,6 +615,7 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
             upload_task_complete_tx,
             shutdown_tx,
             path_resolver,
+            update_from_pow_manager_rx,
         }
     }
 
@@ -660,6 +669,7 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
     }
 
     /// Run one iteration of the reactor loop.
+    #[allow(clippy::cognitive_complexity)] // TODO: Refactor
     async fn run_once(&mut self) -> Result<ShutdownStatus, FatalError> {
         let mut netdir_events = self.dir_provider.events();
 
@@ -798,6 +808,14 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
                     self.upload_all().await?;
                 }
             }
+            update_tp_pow_seed = self.update_from_pow_manager_rx.next().fuse() => {
+                debug!("Update PoW seed for TP!");
+                let Some(time_period) = update_tp_pow_seed else {
+                    return Ok(ShutdownStatus::Terminate);
+                };
+                self.mark_dirty(&time_period);
+                self.upload_all().await?;
+            }
         }
 
         Ok(ShutdownStatus::Continue)
@@ -1296,6 +1314,7 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
     /// The results are received and processed in the main loop of the reactor.
     ///
     /// Returns an error if it fails to spawn a task, or if an internal error occurs.
+    #[allow(clippy::cognitive_complexity)] // TODO #2010: Refactor
     async fn upload_all(&mut self) -> Result<(), FatalError> {
         trace!("starting descriptor upload task...");
 
@@ -1411,6 +1430,7 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
     /// Failed uploads are retried
     /// (see [`upload_descriptor_with_retries`](Reactor::upload_descriptor_with_retries)).
     #[allow(clippy::too_many_arguments)] // TODO: refactor
+    #[allow(clippy::cognitive_complexity)] // TODO: Refactor
     async fn upload_for_time_period(
         hs_dirs: Vec<RelayIds>,
         netdir: &Arc<NetDir>,
@@ -1534,6 +1554,7 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
                                 "building descriptor"
                             );
                             let mut rng = imm.mockable.thread_rng();
+                            let mut key_rng = tor_llcrypto::rng::CautiousRng;
 
                             // We're about to generate a new version of the descriptor,
                             // so let's generate a new revision counter.
@@ -1542,12 +1563,14 @@ impl<R: Runtime, M: Mockable> Reactor<R, M> {
 
                             build_sign(
                                 &imm.keymgr,
+                                &imm.pow_manager,
                                 &config,
                                 authorized_clients.as_deref(),
                                 ipts,
                                 time_period,
                                 revision_counter,
                                 &mut rng,
+                                &mut key_rng,
                                 imm.runtime.wallclock(),
                             )?
                         };
