@@ -161,7 +161,10 @@ define_derive_deftly! {
             s.parse().map_err(|e| E::custom(e))
         }
         fn expecting(&self, f: &mut fmt::Formatter) -> fmt::Result {
-            write!(f, concat!("string representing ", stringify!($tname)))
+            #[allow(clippy::useless_concat)] // False positive
+            {
+                write!(f, concat!("string representing ", stringify!($tname)))
+            }
         }
     }
 }
@@ -199,8 +202,8 @@ pub struct FutureTimestamp {
 /// in human-readable formats, an RFC3339 string with seconds precision and timezone `Z`.)
 #[derive(Debug, Clone, Copy, Eq, PartialEq, Ord, PartialOrd, Hash, derive_more::Display)]
 #[display(
-    fmt = "{}",
-    "humantime::format_rfc3339_seconds(time_t_to_system_time(*time_t))"
+    "{}",
+    humantime::format_rfc3339_seconds(time_t_to_system_time(*time_t))
 )]
 #[derive(Deftly)]
 #[derive_deftly(RawConversions, SerdeStringOrTransparent)]
