@@ -30,18 +30,24 @@
 //! should work for all current future versions of the relay cell crypto design.
 //! The current Tor protocols are instantiated in a `tor1` submodule.
 
+#[cfg(feature = "bench")]
+pub(crate) mod bench_utils;
 #[cfg(feature = "counter-galois-onion")]
 pub(crate) mod cgo;
 pub(crate) mod tor1;
 
 use crate::{Error, Result};
 use derive_deftly::Deftly;
-use tor_cell::{chancell::BoxedCellBody, chancell::ChanCmd};
+use tor_cell::{
+    chancell::{BoxedCellBody, ChanCmd},
+    relaycell::msg::SendmeTag,
+};
 use tor_memquota::derive_deftly_template_HasMemoryCost;
 
 use super::binding::CircuitBinding;
 
 /// Type for the body of a relay cell.
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 #[derive(Clone, derive_more::From, derive_more::Into)]
 pub(crate) struct RelayCellBody(BoxedCellBody);
 
@@ -58,6 +64,7 @@ impl AsMut<[u8]> for RelayCellBody {
 
 /// Represents the ability for one hop of a circuit's cryptographic state to be
 /// initialized from a given seed.
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 pub(crate) trait CryptInit: Sized {
     /// Return the number of bytes that this state will require.
     fn seed_len() -> usize;
@@ -74,6 +81,7 @@ pub(crate) trait CryptInit: Sized {
 /// used by a client to communicate with a single hop on one of its circuits.
 ///
 /// TODO: Maybe we should fold this into CryptInit.
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 pub(crate) trait ClientLayer<F, B>
 where
     F: OutboundClientLayer,
@@ -88,6 +96,7 @@ where
 /// used by a relay to implement a client's circuits.
 ///
 #[allow(dead_code)] // To be used by relays.
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 pub(crate) trait RelayLayer<F, B>
 where
     F: OutboundRelayLayer,
@@ -100,44 +109,48 @@ where
 
 /// Represents a relay's view of the inbound crypto state on a given circuit.
 #[allow(dead_code)] // Relays are not yet implemented.
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 pub(crate) trait InboundRelayLayer {
     /// Prepare a RelayCellBody to be sent towards the client,
     /// and encrypt it.
     ///
     /// Return the authentication tag.
-    fn originate(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> &[u8];
+    fn originate(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> SendmeTag;
     /// Encrypt a RelayCellBody that is moving towards the client.
     fn encrypt_inbound(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody);
 }
 
 /// Represent a relay's view of the outbound crypto state on a given circuit.
 #[allow(dead_code)]
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 pub(crate) trait OutboundRelayLayer {
     /// Decrypt a RelayCellBody that is moving towards the client.
     ///
     /// Return an authentication tag if it is addressed to us.
-    fn decrypt_outbound(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> Option<&[u8]>;
+    fn decrypt_outbound(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> Option<SendmeTag>;
 }
 
 /// A client's view of the cryptographic state shared with a single relay on a
 /// circuit, as used for outbound cells.
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 pub(crate) trait OutboundClientLayer {
     /// Prepare a RelayCellBody to be sent to the relay at this layer, and
     /// encrypt it.
     ///
     /// Return the authentication tag.
-    fn originate_for(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> &[u8];
+    fn originate_for(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> SendmeTag;
     /// Encrypt a RelayCellBody to be decrypted by this layer.
     fn encrypt_outbound(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody);
 }
 
 /// A client's view of the crypto state shared with a single relay on a circuit,
 /// as used for inbound cells.
+#[cfg_attr(feature = "bench", visibility::make(pub))]
 pub(crate) trait InboundClientLayer {
     /// Decrypt a CellBody that passed through this layer.
     ///
     /// Return an authentication tag if this layer is the originator.
-    fn decrypt_inbound(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> Option<&[u8]>;
+    fn decrypt_inbound(&mut self, cmd: ChanCmd, cell: &mut RelayCellBody) -> Option<SendmeTag>;
 }
 
 /// Type to store hop indices on a circuit.
@@ -196,6 +209,7 @@ impl From<HopNum> for usize {
 
 /// A client's view of the cryptographic state for an entire
 /// constructed circuit, as used for sending cells.
+#[cfg_attr(feature = "bench", visibility::make(pub), derive(Default))]
 pub(crate) struct OutboundClientCrypt {
     /// Vector of layers, one for each hop on the circuit, ordered from the
     /// closest hop to the farthest.
@@ -204,18 +218,16 @@ pub(crate) struct OutboundClientCrypt {
 
 /// A client's view of the cryptographic state for an entire
 /// constructed circuit, as used for receiving cells.
+#[cfg_attr(feature = "bench", visibility::make(pub), derive(Default))]
 pub(crate) struct InboundClientCrypt {
     /// Vector of layers, one for each hop on the circuit, ordered from the
     /// closest hop to the farthest.
     layers: Vec<Box<dyn InboundClientLayer + Send>>,
 }
 
-/// The length of the tag that we include (with this algorithm) in an
-/// authenticated SENDME message.
-pub(crate) const SENDME_TAG_LEN: usize = 20;
-
 impl OutboundClientCrypt {
     /// Return a new (empty) OutboundClientCrypt.
+    #[cfg_attr(feature = "bench", visibility::make(pub))]
     pub(crate) fn new() -> Self {
         OutboundClientCrypt { layers: Vec::new() }
     }
@@ -226,12 +238,13 @@ impl OutboundClientCrypt {
     ///
     /// On success, returns a reference to tag that should be expected
     /// for an authenticated SENDME sent in response to this cell.
+    #[cfg_attr(feature = "bench", visibility::make(pub))]
     pub(crate) fn encrypt(
         &mut self,
         cmd: ChanCmd,
         cell: &mut RelayCellBody,
         hop: HopNum,
-    ) -> Result<&[u8; SENDME_TAG_LEN]> {
+    ) -> Result<SendmeTag> {
         let hop: usize = hop.into();
         if hop >= self.layers.len() {
             return Err(Error::NoSuchHop);
@@ -243,7 +256,7 @@ impl OutboundClientCrypt {
         for layer in layers {
             layer.encrypt_outbound(cmd, cell);
         }
-        Ok(tag.try_into().expect("wrong SENDME digest size"))
+        Ok(tag)
     }
 
     /// Add a new layer to this OutboundClientCrypt
@@ -260,6 +273,7 @@ impl OutboundClientCrypt {
 
 impl InboundClientCrypt {
     /// Return a new (empty) InboundClientCrypt.
+    #[cfg_attr(feature = "bench", visibility::make(pub))]
     pub(crate) fn new() -> Self {
         InboundClientCrypt { layers: Vec::new() }
     }
@@ -267,11 +281,12 @@ impl InboundClientCrypt {
     ///
     /// On success, return which hop was the originator of the cell.
     // TODO(nickm): Use a real type for the tag, not just `&[u8]`.
+    #[cfg_attr(feature = "bench", visibility::make(pub))]
     pub(crate) fn decrypt(
         &mut self,
         cmd: ChanCmd,
         cell: &mut RelayCellBody,
-    ) -> Result<(HopNum, &[u8])> {
+    ) -> Result<(HopNum, SendmeTag)> {
         for (hopnum, layer) in self.layers.iter_mut().enumerate() {
             if let Some(tag) = layer.decrypt_inbound(cmd, cell) {
                 let hopnum = HopNum(u8::try_from(hopnum).expect("Somehow > 255 hops"));
@@ -296,15 +311,15 @@ impl InboundClientCrypt {
 }
 
 /// Standard Tor relay crypto, as instantiated for RELAY cells.
-pub(crate) type Tor1RelayCrypto<RCF> =
-    tor1::CryptStatePair<tor_llcrypto::cipher::aes::Aes128Ctr, tor_llcrypto::d::Sha1, RCF>;
+pub(crate) type Tor1RelayCrypto =
+    tor1::CryptStatePair<tor_llcrypto::cipher::aes::Aes128Ctr, tor_llcrypto::d::Sha1>;
 
 /// Standard Tor relay crypto, as instantiated for the HSv3 protocol.
 ///
 /// (The use of SHA3 is ridiculously overkill.)
 #[cfg(feature = "hs-common")]
-pub(crate) type Tor1Hsv3RelayCrypto<RCF> =
-    tor1::CryptStatePair<tor_llcrypto::cipher::aes::Aes256Ctr, tor_llcrypto::d::Sha3_256, RCF>;
+pub(crate) type Tor1Hsv3RelayCrypto =
+    tor1::CryptStatePair<tor_llcrypto::cipher::aes::Aes256Ctr, tor_llcrypto::d::Sha3_256>;
 
 #[cfg(test)]
 mod test {
@@ -326,17 +341,14 @@ mod test {
     use rand::{seq::IndexedRandom as _, RngCore};
     use tor_basic_utils::{test_rng::testing_rng, RngExt as _};
     use tor_bytes::SecretBuf;
-    use tor_cell::relaycell::{
-        RelayCellFields, RelayCellFormat, RelayCellFormatTrait, RelayCellFormatV0,
-    };
+    use tor_cell::relaycell::RelayCellFormat;
 
     pub(crate) fn add_layers(
         cc_out: &mut OutboundClientCrypt,
         cc_in: &mut InboundClientCrypt,
-        // TODO #1067: test other formats
-        pair: Tor1RelayCrypto<RelayCellFormatV0>,
+        pair: Tor1RelayCrypto,
     ) {
-        let (outbound, inbound, _) = pair.split_relay_layer();
+        let (outbound, inbound, _) = pair.split_client_layer();
         cc_out.add_layer(Box::new(outbound));
         cc_in.add_layer(Box::new(inbound));
     }
@@ -365,9 +377,15 @@ mod test {
         assert_eq!(cc_in.n_layers(), 3);
         assert_eq!(cc_out.n_layers(), 3);
 
-        let mut r1 = Tor1RelayCrypto::<RelayCellFormatV0>::construct(KGen::new(seed1)).unwrap();
-        let mut r2 = Tor1RelayCrypto::<RelayCellFormatV0>::construct(KGen::new(seed2)).unwrap();
-        let mut r3 = Tor1RelayCrypto::<RelayCellFormatV0>::construct(KGen::new(seed3)).unwrap();
+        let (mut r1f, mut r1b, _) = Tor1RelayCrypto::construct(KGen::new(seed1))
+            .unwrap()
+            .split_relay_layer();
+        let (mut r2f, mut r2b, _) = Tor1RelayCrypto::construct(KGen::new(seed2))
+            .unwrap()
+            .split_relay_layer();
+        let (mut r3f, mut r3b, _) = Tor1RelayCrypto::construct(KGen::new(seed3))
+            .unwrap()
+            .split_relay_layer();
         let cmd = ChanCmd::RELAY;
 
         let mut rng = testing_rng();
@@ -380,9 +398,9 @@ mod test {
             let mut cell = cell.into();
             let _tag = cc_out.encrypt(cmd, &mut cell, 2.into()).unwrap();
             assert_ne!(&cell.as_ref()[9..], &cell_orig.as_ref()[9..]);
-            assert!(r1.decrypt_outbound(cmd, &mut cell).is_none());
-            assert!(r2.decrypt_outbound(cmd, &mut cell).is_none());
-            assert!(r3.decrypt_outbound(cmd, &mut cell).is_some());
+            assert!(r1f.decrypt_outbound(cmd, &mut cell).is_none());
+            assert!(r2f.decrypt_outbound(cmd, &mut cell).is_none());
+            assert!(r3f.decrypt_outbound(cmd, &mut cell).is_some());
 
             assert_eq!(&cell.as_ref()[9..], &cell_orig.as_ref()[9..]);
 
@@ -393,9 +411,9 @@ mod test {
             cell.copy_from_slice(&cell_orig);
             let mut cell = cell.into();
 
-            r3.originate(cmd, &mut cell);
-            r2.encrypt_inbound(cmd, &mut cell);
-            r1.encrypt_inbound(cmd, &mut cell);
+            r3b.originate(cmd, &mut cell);
+            r2b.encrypt_inbound(cmd, &mut cell);
+            r1b.encrypt_inbound(cmd, &mut cell);
             let (layer, _tag) = cc_in.decrypt(cmd, &mut cell).unwrap();
             assert_eq!(layer, 2.into());
             assert_eq!(&cell.as_ref()[9..], &cell_orig.as_ref()[9..]);
@@ -428,16 +446,16 @@ mod test {
         }
     }
 
-    /// Helper: Clear every field in `cell` that is reserved for cryptography by relay cell
+    /// Helper: Clear every field in the tor1 `cell` that is reserved for cryptography by relay cell
     /// format `version.
     ///
     /// We do this so that we can be sure that the _other_ fields have all been transmitted correctly.
     fn clean_cell_fields(cell: &mut RelayCellBody, format: RelayCellFormat) {
+        use super::tor1;
         match format {
             RelayCellFormat::V0 => {
-                cell.0[<RelayCellFormatV0 as RelayCellFormatTrait>::FIELDS::RECOGNIZED_RANGE]
-                    .fill(0);
-                cell.0[<RelayCellFormatV0 as RelayCellFormatTrait>::FIELDS::DIGEST_RANGE].fill(0);
+                cell.0[tor1::RECOGNIZED_RANGE].fill(0);
+                cell.0[tor1::DIGEST_RANGE].fill(0);
             }
             RelayCellFormat::V1 => {
                 cell.0[0..16].fill(0);
@@ -591,7 +609,7 @@ mod test {
                 .unwrap();
             let hop: u8 = rng.gen_range_checked(0_u8..=2).unwrap();
 
-            let rtag = relays[hop as usize].originate(cmd, &mut cell).to_vec();
+            let rtag = relays[hop as usize].originate(cmd, &mut cell);
             for r_idx in (0..hop.into()).rev() {
                 relays[r_idx as usize].encrypt_inbound(cmd, &mut cell);
             }
@@ -626,12 +644,22 @@ mod test {
         }
     }}
 
-    integration_tests! { tor1(RelayCellFormat::V0, Tor1RelayCrypto<RelayCellFormatV0>, Tor1RelayCrypto<RelayCellFormatV0>) }
+    integration_tests! { tor1(RelayCellFormat::V0, Tor1RelayCrypto, Tor1RelayCrypto) }
     #[cfg(feature = "hs-common")]
-    integration_tests! { tor1_hs(RelayCellFormat::V0, Tor1Hsv3RelayCrypto<RelayCellFormatV0>, Tor1Hsv3RelayCrypto<RelayCellFormatV0>) }
+    integration_tests! { tor1_hs(RelayCellFormat::V0, Tor1Hsv3RelayCrypto, Tor1Hsv3RelayCrypto) }
 
     #[cfg(feature = "counter-galois-onion")]
-    integration_tests! { cgo_aes128(RelayCellFormat::V1, cgo::CryptStatePair<aes::Aes128>, cgo::CryptStatePair<aes::Aes128>) }
+    integration_tests! {
+        cgo_aes128(RelayCellFormat::V1,
+            cgo::CryptStatePair<aes::Aes128Dec, aes::Aes128Enc>,// client
+            cgo::CryptStatePair<aes::Aes128Enc, aes::Aes128Enc> // relay
+        )
+    }
     #[cfg(feature = "counter-galois-onion")]
-    integration_tests! { cgo_aes256(RelayCellFormat::V1, cgo::CryptStatePair<aes::Aes256>, cgo::CryptStatePair<aes::Aes256>) }
+    integration_tests! {
+        cgo_aes256(RelayCellFormat::V1,
+            cgo::CryptStatePair<aes::Aes256Dec, aes::Aes256Enc>,// client
+            cgo::CryptStatePair<aes::Aes256Enc, aes::Aes256Enc> // relay
+        )
+    }
 }
