@@ -3,7 +3,7 @@
 //! A stream is an anonymized conversation; multiple streams can be
 //! multiplexed over a single circuit.
 //!
-//! To create a stream, use [crate::circuit::ClientCirc::begin_stream].
+//! To create a stream, use [crate::tunnel::circuit::ClientCirc::begin_stream].
 //!
 //! # Limitations
 //!
@@ -32,7 +32,7 @@ pub use incoming::{
     IncomingStreamRequestDisposition, IncomingStreamRequestFilter,
 };
 pub use params::StreamParameters;
-pub use raw::StreamReader;
+pub use raw::StreamReceiver;
 pub use resolve::ResolveStream;
 pub(crate) use {data::DataCmdChecker, resolve::ResolveCmdChecker};
 
