@@ -30,7 +30,7 @@ use crate::stream::{IncomingStreamRequest, IncomingStreamRequestFilter};
 use crate::tunnel::circuit::celltypes::ClientCircChanMsg;
 use crate::tunnel::circuit::unique_id::UniqId;
 use crate::tunnel::circuit::CircuitRxReceiver;
-use crate::tunnel::{streammap, HopLocation, TargetHop};
+use crate::tunnel::{streammap, HopLocation, TargetHop, TunnelId, TunnelScopedCircId};
 use crate::util::err::ReactorError;
 use crate::util::skew::ClockSkew;
 use crate::{Error, Result};
@@ -60,9 +60,9 @@ use derive_deftly::Deftly;
 use derive_more::From;
 use tor_cell::chancell::CircId;
 use tor_llcrypto::pk;
+use tor_memquota::derive_deftly_template_HasMemoryCost;
 use tor_memquota::mq_queue::{self, MpscSpec};
-use tor_memquota::{derive_deftly_template_HasMemoryCost, memory_cost_structural_copy};
-use tracing::trace;
+use tracing::{info, trace, warn};
 
 use super::circuit::{MutableState, TunnelMutableState};
 
@@ -134,7 +134,7 @@ impl Default for CloseStreamBehavior {
     }
 }
 
-// TODO(conflux): the RunOnceCmd/RunOnceCmdInner/CircuitCmd/CircuitAction enum
+// TODO: the RunOnceCmd/RunOnceCmdInner/CircuitCmd/CircuitAction enum
 // proliferation is a bit bothersome, but unavoidable with the current design.
 //
 // We should consider getting rid of some of these enums (if possible),
@@ -166,7 +166,7 @@ enum RunOnceCmdInner {
     /// Send a RELAY cell.
     Send {
         /// The leg the cell should be sent on.
-        leg: LegId,
+        leg: UniqId,
         /// The cell to send.
         cell: SendRelayCell,
         /// A channel for sending completion notifications.
@@ -190,7 +190,7 @@ enum RunOnceCmdInner {
     /// Handle a SENDME message.
     HandleSendMe {
         /// The leg the SENDME was received on.
-        leg: LegId,
+        leg: UniqId,
         /// The hop number.
         hop: HopNum,
         /// The SENDME message to handle.
@@ -207,7 +207,7 @@ enum RunOnceCmdInner {
         /// caller.
         hop: HopLocation,
         /// The circuit leg to begin the stream on.
-        leg: LegId,
+        leg: UniqId,
         /// Oneshot channel to notify on completion, with the allocated stream ID.
         done: ReactorResultChannel<(StreamId, HopLocation, RelayCellFormat)>,
     },
@@ -232,7 +232,7 @@ enum RunOnceCmdInner {
     /// Remove a circuit leg from the conflux set.
     RemoveLeg {
         /// The circuit leg to remove.
-        leg: LegId,
+        leg: UniqId,
         /// The reason for removal.
         ///
         /// This is only used for conflux circuits that get removed
@@ -257,7 +257,7 @@ enum RunOnceCmdInner {
     ConfluxHandshakeComplete {
         /// The circuit leg that has completed the handshake,
         /// This is the leg the cell should be sent on.
-        leg: LegId,
+        leg: UniqId,
         /// The cell to send.
         cell: SendRelayCell,
     },
@@ -274,7 +274,7 @@ enum RunOnceCmdInner {
     #[cfg(feature = "conflux")]
     Enqueue {
         /// The leg the entry originated from.
-        leg: LegId,
+        leg: UniqId,
         /// The out-of-order message.
         msg: OooRelayMsg,
     },
@@ -283,46 +283,35 @@ enum RunOnceCmdInner {
 }
 
 impl RunOnceCmdInner {
-    /// Create a [`RunOnceCmdInner`] out of a [`CircuitCmd`] and [`LegIdKey`].
-    fn from_circuit_cmd(leg: LegIdKey, cmd: CircuitCmd) -> Self {
+    /// Create a [`RunOnceCmdInner`] out of a [`CircuitCmd`] and [`UniqId`].
+    fn from_circuit_cmd(leg: UniqId, cmd: CircuitCmd) -> Self {
         match cmd {
             CircuitCmd::Send(cell) => Self::Send {
-                leg: LegId(leg),
+                leg,
                 cell,
                 done: None,
             },
-            CircuitCmd::HandleSendMe { hop, sendme } => Self::HandleSendMe {
-                leg: LegId(leg),
-                hop,
-                sendme,
-            },
+            CircuitCmd::HandleSendMe { hop, sendme } => Self::HandleSendMe { leg, hop, sendme },
             CircuitCmd::CloseStream {
                 hop,
                 sid,
                 behav,
                 reason,
             } => Self::CloseStream {
-                hop: HopLocation::Hop((LegId(leg), hop)),
+                hop: HopLocation::Hop((leg, hop)),
                 sid,
                 behav,
                 reason,
                 done: None,
             },
             #[cfg(feature = "conflux")]
-            CircuitCmd::ConfluxRemove(reason) => Self::RemoveLeg {
-                leg: LegId(leg),
-                reason,
-            },
+            CircuitCmd::ConfluxRemove(reason) => Self::RemoveLeg { leg, reason },
             #[cfg(feature = "conflux")]
-            CircuitCmd::ConfluxHandshakeComplete(cell) => Self::ConfluxHandshakeComplete {
-                leg: LegId(leg),
-                cell,
-            },
+            CircuitCmd::ConfluxHandshakeComplete(cell) => {
+                Self::ConfluxHandshakeComplete { leg, cell }
+            }
             #[cfg(feature = "conflux")]
-            CircuitCmd::Enqueue(msg) => Self::Enqueue {
-                leg: LegId(leg),
-                msg,
-            },
+            CircuitCmd::Enqueue(msg) => Self::Enqueue { leg, msg },
             CircuitCmd::CleanShutdown => Self::CleanShutdown,
         }
     }
@@ -349,7 +338,7 @@ enum CircuitAction {
     /// Run a single `CircuitCmd` command.
     RunCmd {
         /// The unique identifier of the circuit leg to run the command on
-        leg: LegIdKey,
+        leg: UniqId,
         /// The command to run.
         cmd: CircuitCmd,
     },
@@ -358,7 +347,7 @@ enum CircuitAction {
     /// Handle an input message.
     HandleCell {
         /// The unique identifier of the circuit leg the message was received on.
-        leg: LegIdKey,
+        leg: UniqId,
         /// The message to handle.
         cell: ClientCircChanMsg,
     },
@@ -374,7 +363,7 @@ enum CircuitAction {
     /// See the [`ConfluxSet::remove`] docs for more on the exact behavior of this command.
     RemoveLeg {
         /// The leg to remove.
-        leg: LegIdKey,
+        leg: UniqId,
         /// The reason for removal.
         ///
         /// This is only used for conflux circuits that get removed
@@ -388,16 +377,19 @@ enum CircuitAction {
 }
 
 /// The reason for removing a circuit leg from the conflux set.
-#[derive(Debug)]
+#[derive(Debug, derive_more::Display)]
 enum RemoveLegReason {
     /// The conflux handshake timed out.
     ///
     /// On the client-side, this means we didn't receive
     /// the CONFLUX_LINKED response in time.
+    #[display("conflux handshake timed out")]
     ConfluxHandshakeTimeout,
     /// An error occurred during conflux handshake.
+    #[display("{}", _0)]
     ConfluxHandshakeErr(Error),
     /// The channel was closed.
+    #[display("channel closed")]
     ChannelClosed,
 }
 
@@ -414,7 +406,7 @@ enum RemoveLegReason {
 pub(crate) trait MetaCellHandler: Send {
     /// The hop we're expecting the message to come from. This is compared against the hop
     /// from which we actually receive messages, and an error is thrown if the two don't match.
-    fn expected_hop(&self) -> HopNum;
+    fn expected_hop(&self) -> HopLocation;
     /// Called when the message we were waiting for arrives.
     ///
     /// Gets a copy of the `Reactor` in order to do anything it likes there.
@@ -445,35 +437,6 @@ pub(crate) enum MetaCellDisposition {
     // But right now we don't need that.
 }
 
-/// A unique identifier for a circuit leg.
-///
-/// After the circuit is torn down, its `LegId` becomes invalid.
-/// The same `LegId` won't be reused for a future circuit.
-//
-// TODO(#1857): make this pub
-#[allow(unused)]
-#[derive(Copy, Clone, Debug, Eq, PartialEq, Deftly)]
-#[derive_deftly(HasMemoryCost)]
-pub(crate) struct LegId(pub(crate) LegIdKey);
-
-// TODO(#1999): can we use `UniqId` as the key instead of this newtype?
-//
-// See https://gitlab.torproject.org/tpo/core/arti/-/merge_requests/2996#note_3199069
-slotmap_careful::new_key_type! {
-    /// A key type for the circuit leg slotmap
-    ///
-    /// See [`LegId`].
-    pub(crate) struct LegIdKey;
-}
-
-impl From<LegIdKey> for LegId {
-    fn from(leg_id: LegIdKey) -> Self {
-        LegId(leg_id)
-    }
-}
-
-memory_cost_structural_copy!(LegIdKey);
-
 /// Unwrap the specified [`Option`], returning a [`ReactorError::Shutdown`] if it is `None`.
 ///
 /// This is a macro instead of a function to work around borrowck errors
@@ -482,7 +445,11 @@ macro_rules! unwrap_or_shutdown {
     ($self:expr, $res:expr, $reason:expr) => {{
         match $res {
             None => {
-                trace!("{}: reactor shutdown due to {}", $self.unique_id, $reason);
+                trace!(
+                    tunnel_id = %$self.tunnel_id,
+                    reason = %$reason,
+                    "reactor shutdown"
+                );
                 Err(ReactorError::Shutdown)
             }
             Some(v) => Ok(v),
@@ -532,8 +499,8 @@ pub struct Reactor {
     // TODO(conflux): document all the reasons why the reactor might
     // chose to tear down a circuit or tunnel (timeouts, protocol violations, etc.)
     circuits: ConfluxSet,
-    /// An identifier for logging about this reactor's circuit.
-    unique_id: UniqId,
+    /// An identifier for logging about this tunnel reactor.
+    tunnel_id: TunnelId,
     /// Handlers, shared with `Circuit`.
     cell_handlers: CellHandlers,
     /// The time provider, used for conflux handshake timeouts.
@@ -570,7 +537,7 @@ struct ConfluxHandshakeCtx {
 #[cfg(feature = "conflux")]
 struct ConfluxHeapEntry {
     /// The leg id this message came from.
-    leg_id: LegId,
+    leg_id: UniqId,
     /// The out of order message
     msg: OooRelayMsg,
 }
@@ -624,9 +591,7 @@ pub(crate) struct StreamReqInfo {
     //
     // TODO: For onion services, we might be able to enforce the HopNum earlier: we would never accept an
     // incoming stream request from two separate hops.  (There is only one that's valid.)
-    pub(crate) hop_num: HopNum,
-    /// The [`LegId`] of the circuit the request came on.
-    pub(crate) leg: LegId,
+    pub(crate) hop: HopLocation,
     /// The format which must be used with this stream to encode messages.
     #[deftly(has_memory_cost(indirect_size = "0"))]
     pub(crate) relay_cell_format: RelayCellFormat,
@@ -681,6 +646,7 @@ impl Reactor {
         oneshot::Receiver<void::Void>,
         Arc<TunnelMutableState>,
     ) {
+        let tunnel_id = TunnelId::next();
         let (control_tx, control_rx) = mpsc::unbounded();
         let (command_tx, command_rx) = mpsc::unbounded();
         let mutable = Arc::new(MutableState::default());
@@ -693,7 +659,9 @@ impl Reactor {
             incoming_stream_req_handler: None,
         };
 
+        let unique_id = TunnelScopedCircId::new(tunnel_id, unique_id);
         let circuit_leg = Circuit::new(
+            runtime.clone(),
             channel,
             channel_id,
             unique_id,
@@ -702,14 +670,14 @@ impl Reactor {
             Arc::clone(&mutable),
         );
 
-        let (circuits, mutable) = ConfluxSet::new(circuit_leg);
+        let (circuits, mutable) = ConfluxSet::new(tunnel_id, circuit_leg);
 
         let reactor = Reactor {
             circuits,
             control: control_rx,
             command: command_rx,
             reactor_closed_tx,
-            unique_id,
+            tunnel_id,
             cell_handlers,
             runtime,
             #[cfg(feature = "conflux")]
@@ -727,7 +695,7 @@ impl Reactor {
     /// Once this method returns, the circuit is dead and cannot be
     /// used again.
     pub async fn run(mut self) -> Result<()> {
-        trace!("{}: Running circuit reactor", self.unique_id);
+        trace!(tunnel_id = %self.tunnel_id, "Running tunnel reactor");
         let result: Result<()> = loop {
             match self.run_once().await {
                 Ok(()) => (),
@@ -735,7 +703,7 @@ impl Reactor {
                 Err(ReactorError::Err(e)) => break Err(e),
             }
         };
-        trace!("{}: Circuit reactor stopped: {:?}", self.unique_id, result);
+        trace!(tunnel_id = %self.tunnel_id, "Tunnel reactor stopped: {:?}", result);
         result
     }
 
@@ -743,12 +711,10 @@ impl Reactor {
     /// processes one cell or control message.
     async fn run_once(&mut self) -> StdResult<(), ReactorError> {
         // If all the circuits are closed, shut down the reactor
-        //
-        // TODO(conflux): we might need to rethink this behavior
         if self.circuits.is_empty() {
             trace!(
-                "{}: Circuit reactor shutting down: all circuits have closed",
-                self.unique_id
+                tunnel_id = %self.tunnel_id,
+                "Tunnel reactor shutting down: all circuits have closed",
             );
 
             return Err(ReactorError::Shutdown);
@@ -759,7 +725,7 @@ impl Reactor {
         let single_path_with_hops = self
             .circuits
             .single_leg_mut()
-            .is_ok_and(|(_id, leg)| !leg.has_hops());
+            .is_ok_and(|leg| !leg.has_hops());
         if single_path_with_hops {
             self.wait_for_create().await?;
 
@@ -790,7 +756,7 @@ impl Reactor {
                 let msg = unwrap_or_shutdown!(self, ret, "control drop")?;
                 CircuitAction::HandleControl(msg)
             },
-            res = self.circuits.next_circ_action(&self.runtime).fuse() => res?,
+            res = self.circuits.next_circ_action(&self.runtime)?.fuse() => res?,
         };
 
         let cmd = match action {
@@ -803,16 +769,16 @@ impl Reactor {
             CircuitAction::HandleCell { leg, cell } => {
                 let circ = self
                     .circuits
-                    .leg_mut(LegId(leg))
+                    .leg_mut(leg)
                     .ok_or_else(|| internal!("the circuit leg we just had disappeared?!"))?;
 
-                let circ_cmds = circ.handle_cell(&mut self.cell_handlers, leg.into(), cell)?;
+                let circ_cmds = circ.handle_cell(&mut self.cell_handlers, leg, cell)?;
                 if circ_cmds.is_empty() {
                     None
                 } else {
-                    // TODO(conflux): we return RunOnceCmd::Multiple even if there's a single command.
+                    // TODO: we return RunOnceCmd::Multiple even if there's a single command.
                     //
-                    // See the TODO(conflux) on `Circuit::handle_cell`.
+                    // See the TODO on `Circuit::handle_cell`.
                     let cmd = RunOnceCmd::Multiple(
                         circ_cmds
                             .into_iter()
@@ -823,32 +789,15 @@ impl Reactor {
                     Some(cmd)
                 }
             }
-            CircuitAction::RemoveLeg { leg, reason } => Some(
-                RunOnceCmdInner::RemoveLeg {
-                    leg: LegId(leg),
-                    reason,
-                }
-                .into(),
-            ),
+            CircuitAction::RemoveLeg { leg, reason } => {
+                Some(RunOnceCmdInner::RemoveLeg { leg, reason }.into())
+            }
         };
 
         if let Some(cmd) = cmd {
             self.handle_run_once_cmd(cmd).await?;
         }
 
-        // Check if it's time to switch our primary leg.
-        //
-        // TODO(conflux): this only be done just before sending a cell.
-        //
-        // See https://gitlab.torproject.org/tpo/core/arti/-/merge_requests/2946#note_3192013
-        #[cfg(feature = "conflux")]
-        if let Some(switch_cell) = self.circuits.maybe_update_primary_leg()? {
-            self.circuits
-                .primary_leg_mut()?
-                .send_relay_cell(switch_cell)
-                .await?;
-        }
-
         Ok(())
     }
 
@@ -880,7 +829,7 @@ impl Reactor {
                     entry.msg.msg,
                 )?
                 .map(|cmd| {
-                    RunOnceCmd::Single(RunOnceCmdInner::from_circuit_cmd(entry.leg_id.0, cmd))
+                    RunOnceCmd::Single(RunOnceCmdInner::from_circuit_cmd(entry.leg_id, cmd))
                 });
 
             if let Some(cmd) = cmd {
@@ -918,11 +867,7 @@ impl Reactor {
         match cmd {
             RunOnceCmdInner::Send { leg, cell, done } => {
                 // TODO: check the cc window
-                let leg = self
-                    .circuits
-                    .leg_mut(leg)
-                    .ok_or_else(|| internal!("leg disappeared?!"))?;
-                let res = leg.send_relay_cell(cell).await;
+                let res = self.circuits.send_relay_cell_on_leg(cell, Some(leg)).await;
                 if let Some(done) = done {
                     // Don't care if the receiver goes away
                     let _ = done.send(res.clone());
@@ -937,7 +882,7 @@ impl Reactor {
                 match cell {
                     Ok(Some(cell)) => {
                         // TODO(conflux): let the RunOnceCmdInner specify which leg to send the cell on
-                        let outcome = self.circuits.primary_leg_mut()?.send_relay_cell(cell).await;
+                        let outcome = self.circuits.send_relay_cell_on_leg(cell, None).await;
                         // don't care if receiver goes away.
                         let _ = done.send(outcome.clone());
                         outcome?;
@@ -961,17 +906,18 @@ impl Reactor {
             } => {
                 match cell {
                     Ok((cell, stream_id)) => {
-                        let leg = self
+                        let circ = self
                             .circuits
                             .leg_mut(leg)
                             .ok_or_else(|| internal!("leg disappeared?!"))?;
                         let cell_hop = cell.hop;
-                        let relay_format = leg
+                        let relay_format = circ
                             .hop_mut(cell_hop)
                             // TODO: Is this the right error type here? Or should there be a "HopDisappeared"?
                             .ok_or(Error::NoSuchHop)?
                             .relay_cell_format();
-                        let outcome = leg.send_relay_cell(cell).await;
+
+                        let outcome = self.circuits.send_relay_cell_on_leg(cell, Some(leg)).await;
                         // don't care if receiver goes away.
                         let _ = done.send(outcome.clone().map(|_| (stream_id, hop, relay_format)));
                         outcome?;
@@ -1033,20 +979,19 @@ impl Reactor {
                 leg.handle_sendme(hop, sendme, signals)?;
             }
             RunOnceCmdInner::FirstHopClockSkew { answer } => {
-                let res = self
-                    .circuits
-                    .single_leg_mut()
-                    .map(|(_id, leg)| leg.clock_skew());
+                let res = self.circuits.single_leg_mut().map(|leg| leg.clock_skew());
 
                 // don't care if the sender goes away
                 let _ = answer.send(res.map_err(Into::into));
             }
             RunOnceCmdInner::CleanShutdown => {
-                trace!("{}: reactor shutdown due to handled cell", self.unique_id);
+                trace!(tunnel_id = %self.tunnel_id, "reactor shutdown due to handled cell");
                 return Err(ReactorError::Shutdown);
             }
             RunOnceCmdInner::RemoveLeg { leg, reason } => {
-                let circ = self.circuits.remove(leg.0)?;
+                warn!(tunnel_id = %self.tunnel_id, reason = %reason, "removing circuit leg");
+
+                let circ = self.circuits.remove(leg)?;
                 let is_conflux_pending = circ.is_conflux_pending();
 
                 // Drop the removed leg. This will cause it to close if it's not already closed.
@@ -1058,13 +1003,30 @@ impl Reactor {
 
                 #[cfg(feature = "conflux")]
                 if is_conflux_pending {
-                    let error = match reason {
-                        RemoveLegReason::ConfluxHandshakeTimeout => ConfluxHandshakeError::Timeout,
-                        RemoveLegReason::ConfluxHandshakeErr(e) => ConfluxHandshakeError::Link(e),
-                        RemoveLegReason::ChannelClosed => ConfluxHandshakeError::ChannelClosed,
+                    let (error, proto_violation): (_, Option<Error>) = match &reason {
+                        RemoveLegReason::ConfluxHandshakeTimeout => {
+                            (ConfluxHandshakeError::Timeout, None)
+                        }
+                        RemoveLegReason::ConfluxHandshakeErr(e) => {
+                            (ConfluxHandshakeError::Link(e.clone()), Some(e.clone()))
+                        }
+                        RemoveLegReason::ChannelClosed => {
+                            (ConfluxHandshakeError::ChannelClosed, None)
+                        }
                     };
 
-                    self.note_conflux_handshake_result(Err(error))?;
+                    self.note_conflux_handshake_result(Err(error), proto_violation.is_some())?;
+
+                    if let Some(e) = proto_violation {
+                        // TODO: make warn_report support structured logging
+                        tor_error::warn_report!(
+                            e,
+                            "{}: Malformed conflux handshake, tearing down tunnel",
+                            self.tunnel_id
+                        );
+
+                        return Err(e.into());
+                    }
                 }
             }
             #[cfg(feature = "conflux")]
@@ -1078,13 +1040,9 @@ impl Reactor {
                 // down anyway. OTOH, marking the handshake as complete slightly early
                 // means that on the happy path, the circuit is marked as usable sooner,
                 // instead of blocking on the sending of the LINKED_ACK.
-                self.note_conflux_handshake_result(Ok(()))?;
+                self.note_conflux_handshake_result(Ok(()), false)?;
 
-                let leg = self
-                    .circuits
-                    .leg_mut(leg)
-                    .ok_or_else(|| internal!("leg disappeared?!"))?;
-                let res = leg.send_relay_cell(cell).await;
+                let res = self.circuits.send_relay_cell_on_leg(cell, Some(leg)).await;
 
                 res?;
             }
@@ -1120,11 +1078,12 @@ impl Reactor {
                         relay_cell_format: format,
                         fwd_lasthop,
                         rev_lasthop,
+                        peer_id,
                         params,
                         done,
                     } => {
-                        let (_id, leg) = self.circuits.single_leg_mut()?;
-                        leg.handle_add_fake_hop(format, fwd_lasthop, rev_lasthop, &params, done);
+                        let leg = self.circuits.single_leg_mut()?;
+                        leg.handle_add_fake_hop(format, fwd_lasthop, rev_lasthop, peer_id, &params, done);
                         return Ok(())
                     },
                     _ => {
@@ -1140,13 +1099,13 @@ impl Reactor {
             CtrlMsg::Create {
                 recv_created,
                 handshake,
-                mut params,
+                settings,
                 done,
             } => {
                 // TODO(conflux): instead of crashing the reactor, it might be better
                 // to send the error via the done channel instead
-                let (_id, leg) = self.circuits.single_leg_mut()?;
-                leg.handle_create(recv_created, handshake, &mut params, done)
+                let leg = self.circuits.single_leg_mut()?;
+                leg.handle_create(recv_created, handshake, settings, done)
                     .await
             }
             _ => {
@@ -1166,6 +1125,7 @@ impl Reactor {
     fn note_conflux_handshake_result(
         &mut self,
         res: StdResult<(), ConfluxHandshakeError>,
+        reactor_is_closing: bool,
     ) -> StdResult<(), ReactorError> {
         let tunnel_complete = match self.conflux_hs_ctx.as_mut() {
             Some(conflux_ctx) => {
@@ -1178,11 +1138,19 @@ impl Reactor {
             }
         };
 
-        if tunnel_complete {
+        if tunnel_complete || reactor_is_closing {
             // Time to remove the conflux handshake context
             // and extract the results we have collected
             let conflux_ctx = self.conflux_hs_ctx.take().expect("context disappeared?!");
 
+            let success_count = conflux_ctx.results.iter().filter(|res| res.is_ok()).count();
+            let leg_count = conflux_ctx.results.len();
+
+            info!(
+                tunnel_id = %self.tunnel_id,
+                "conflux tunnel ready ({success_count}/{leg_count} circuits successfully linked)",
+            );
+
             send_conflux_outcome(conflux_ctx.answer, Ok(conflux_ctx.results))?;
 
             // We don't expect to receive any more handshake results,
@@ -1207,8 +1175,13 @@ impl Reactor {
                     .as_ref()
                     .or(handlers.meta_handler.as_ref())
                     .ok_or_else(|| internal!("tried to use an ended Conversation"))?;
+                // We should always have a precise HopLocation here so this should never fails but
+                // in case we have a ::JointPoint, we'll notice.
+                let hop = handler.expected_hop().hop_num().ok_or(bad_api_usage!(
+                    "MsgHandler doesn't have a precise HopLocation"
+                ))?;
                 Ok::<_, crate::Error>(SendRelayCell {
-                    hop: handler.expected_hop(),
+                    hop,
                     early: false,
                     cell: msg,
                 })
@@ -1225,8 +1198,8 @@ impl Reactor {
     /// Handle a shutdown request.
     fn handle_shutdown(&self) -> StdResult<Option<RunOnceCmdInner>, ReactorError> {
         trace!(
-            "{}: reactor shutdown due to explicit request",
-            self.unique_id
+            tunnel_id = %self.tunnel_id,
+            "reactor shutdown due to explicit request",
         );
 
         Err(ReactorError::Shutdown)
@@ -1265,13 +1238,14 @@ impl Reactor {
         match hop {
             TargetHop::Hop(hop) => Ok(hop),
             TargetHop::LastHop => {
-                if let Ok((leg_id, leg)) = self.circuits.single_leg() {
+                if let Ok(leg) = self.circuits.single_leg() {
+                    let leg_id = leg.unique_id();
                     // single-path tunnel
                     let hop = leg.last_hop_num().ok_or(NoHopsBuiltError)?;
                     Ok(HopLocation::Hop((leg_id, hop)))
                 } else if !self.circuits.is_empty() {
                     // multi-path tunnel
-                    return Ok(HopLocation::JoinPoint);
+                    Ok(HopLocation::JoinPoint)
                 } else {
                     // no legs
                     Err(NoHopsBuiltError)
@@ -1280,14 +1254,14 @@ impl Reactor {
         }
     }
 
-    /// Resolves a [`HopLocation`] to a [`LegId`] and [`HopNum`].
+    /// Resolves a [`HopLocation`] to a [`UniqId`] and [`HopNum`].
     ///
     /// After resolving a `HopLocation::JoinPoint`,
-    /// the [`LegId`] and [`HopNum`] can become stale if the primary leg changes.
+    /// the [`UniqId`] and [`HopNum`] can become stale if the primary leg changes.
     ///
-    /// You should try to only resolve to a specific [`LegId`] and [`HopNum`] immediately before you
+    /// You should try to only resolve to a specific [`UniqId`] and [`HopNum`] immediately before you
     /// need them,
-    /// and you should not hold on to the resolved [`LegId`] and [`HopNum`] between reactor
+    /// and you should not hold on to the resolved [`UniqId`] and [`HopNum`] between reactor
     /// iterations as the primary leg may change from one iteration to the next.
     ///
     /// Returns [`NoJoinPointError`] if trying to resolve `HopLocation::JoinPoint`
@@ -1295,7 +1269,7 @@ impl Reactor {
     fn resolve_hop_location(
         &self,
         hop: HopLocation,
-    ) -> StdResult<(LegId, HopNum), NoJoinPointError> {
+    ) -> StdResult<(UniqId, HopNum), NoJoinPointError> {
         match hop {
             HopLocation::Hop((leg_id, hop_num)) => Ok((leg_id, hop_num)),
             HopLocation::JoinPoint => {
@@ -1309,10 +1283,23 @@ impl Reactor {
         }
     }
 
+    /// Resolve a [`TargetHop`] directly into a [`UniqId`] and [`HopNum`].
+    ///
+    /// This is a helper function that basically calls both resolve_target_hop and
+    /// resolve_hop_location back to back.
+    ///
+    /// It returns None on failure to resolve meaning that if you want more detailed error on why
+    /// it failed, explicitly use the resolve_hop_location() and resolve_target_hop() functions.
+    pub(crate) fn target_hop_to_hopnum_id(&self, hop: TargetHop) -> Option<(UniqId, HopNum)> {
+        self.resolve_target_hop(hop)
+            .ok()
+            .and_then(|resolved| self.resolve_hop_location(resolved).ok())
+    }
+
     /// Does congestion control use stream SENDMEs for the given hop?
     ///
     /// Returns `None` if either the `leg` or `hop` don't exist.
-    fn uses_stream_sendme(&self, leg: LegId, hop: HopNum) -> Option<bool> {
+    fn uses_stream_sendme(&self, leg: UniqId, hop: HopNum) -> Option<bool> {
         self.circuits.uses_stream_sendme(leg, hop)
     }
 
@@ -1332,6 +1319,8 @@ impl Reactor {
         circuits: Vec<Circuit>,
         answer: ConfluxLinkResultChannel,
     ) -> StdResult<(), ReactorError> {
+        use tor_error::warn_report;
+
         if self.conflux_hs_ctx.is_some() {
             let err = internal!("conflux linking already in progress");
             send_conflux_outcome(answer, Err(err.into()))?;
@@ -1339,20 +1328,27 @@ impl Reactor {
             return Ok(());
         }
 
-        let num_legs = circuits.len();
+        let unlinked_legs = self.circuits.num_unlinked();
+
+        // We need to send the LINK cell on each of the new circuits
+        // and on each of the existing, unlinked legs from self.circuits.
+        //
+        // In reality, there can only be one such circuit
+        // (the "initial" one from the previously single-path tunnel),
+        // because any circuits that to complete the conflux handshake
+        // get removed from the set.
+        let num_legs = circuits.len() + unlinked_legs;
 
         // Note: add_legs validates `circuits`
         let res = async {
             self.circuits.add_legs(circuits, &self.runtime)?;
-
-            // TODO(conflux): check if we negotiated prop324 cc on *all* circuits,
-            // returning an error if not?
-
             self.circuits.link_circuits(&self.runtime).await
         }
         .await;
 
         if let Err(e) = res {
+            warn_report!(e, "Failed to link conflux circuits");
+
             send_conflux_outcome(answer, Err(e))?;
         } else {
             // Save the channel, to notify the user of completion.
