@@ -5,22 +5,25 @@ use super::{
     CircuitHandshake, CloseStreamBehavior, MetaCellHandler, Reactor, ReactorResultChannel,
     RunOnceCmdInner, SendRelayCell,
 };
+use crate::circuit::HopSettings;
 use crate::crypto::binding::CircuitBinding;
-use crate::crypto::cell::{HopNum, InboundClientLayer, OutboundClientLayer, Tor1RelayCrypto};
+use crate::crypto::cell::{InboundClientLayer, OutboundClientLayer, Tor1RelayCrypto};
 use crate::crypto::handshake::ntor_v3::{NtorV3Client, NtorV3PublicKey};
 use crate::stream::AnyCmdChecker;
 use crate::tunnel::circuit::celltypes::CreateResponse;
-use crate::tunnel::circuit::{path, CircParameters};
-use crate::tunnel::reactor::circuit::circ_extensions_from_params;
-use crate::tunnel::reactor::{NtorClient, ReactorError};
+use crate::tunnel::circuit::path;
+use crate::tunnel::reactor::circuit::circ_extensions_from_settings;
+use crate::tunnel::reactor::{NoJoinPointError, NtorClient, ReactorError};
 use crate::tunnel::{streammap, HopLocation, TargetHop};
 use crate::util::skew::ClockSkew;
 use crate::Result;
+#[cfg(test)]
+use crate::{circuit::CircParameters, circuit::UniqId, crypto::cell::HopNum};
 use tor_cell::chancell::msg::HandshakeType;
 use tor_cell::relaycell::msg::{AnyRelayMsg, Sendme};
 use tor_cell::relaycell::{AnyRelayMsgOuter, RelayCellFormat, StreamId, UnparsedRelayMsg};
-use tor_error::{bad_api_usage, into_bad_api_usage, Bug};
-use tracing::trace;
+use tor_error::{bad_api_usage, internal, into_bad_api_usage, warn_report, Bug};
+use tracing::{debug, trace};
 #[cfg(feature = "hs-service")]
 use {
     super::StreamReqSender, crate::stream::IncomingStreamRequestFilter,
@@ -58,7 +61,7 @@ pub(crate) enum CtrlMsg {
         /// The handshake type to use for the first hop.
         handshake: CircuitHandshake,
         /// Other parameters relevant for circuit creation.
-        params: CircParameters,
+        settings: HopSettings,
         /// Oneshot channel to notify on completion.
         done: ReactorResultChannel<()>,
     },
@@ -72,8 +75,8 @@ pub(crate) enum CtrlMsg {
         public_key: NtorPublicKey,
         /// Information about how to connect to the relay we're extending to.
         linkspecs: Vec<EncodedLinkSpec>,
-        /// Other parameters relevant for circuit extension.
-        params: CircParameters,
+        /// Other parameters we are negotiating.
+        settings: HopSettings,
         /// Oneshot channel to notify on completion.
         done: ReactorResultChannel<()>,
     },
@@ -87,8 +90,8 @@ pub(crate) enum CtrlMsg {
         public_key: NtorV3PublicKey,
         /// Information about how to connect to the relay we're extending to.
         linkspecs: Vec<EncodedLinkSpec>,
-        /// Other parameters relevant for circuit extension.
-        params: CircParameters,
+        /// Other parameters we are negotiating.
+        settings: HopSettings,
         /// Oneshot channel to notify on completion.
         done: ReactorResultChannel<()>,
     },
@@ -136,7 +139,7 @@ pub(crate) enum CtrlMsg {
     #[cfg(feature = "send-control-msg")]
     SendMsg {
         /// The hop to receive this message.
-        hop_num: HopNum,
+        hop: TargetHop,
         /// The message to send.
         msg: AnyRelayMsg,
         /// A sender that we use to tell the caller that the message was sent
@@ -164,8 +167,6 @@ pub(crate) enum CtrlMsg {
         stream_id: StreamId,
         /// The hop number the stream is on.
         hop: HopLocation,
-        /// A sender that we use to tell the caller that the SENDME was sent.
-        sender: oneshot::Sender<Result<()>>,
     },
     /// Get the clock skew claimed by the first hop of the circuit.
     FirstHopClockSkew {
@@ -218,11 +219,18 @@ pub(crate) enum CtrlCmd {
             Box<dyn InboundClientLayer + Send>,
             Option<CircuitBinding>,
         ),
-        /// A set of parameters used to configure this hop.
-        params: CircParameters,
+        /// A set of parameters to negotiate with this hop.
+        settings: HopSettings,
         /// Oneshot channel to notify on completion.
         done: ReactorResultChannel<()>,
     },
+    /// Resolve a given [`TargetHop`] into a precise [`HopLocation`].
+    ResolveTargetHop {
+        /// The target hop to resolve.
+        hop: TargetHop,
+        /// Oneshot channel to notify on completion.
+        done: ReactorResultChannel<HopLocation>,
+    },
     /// Begin accepting streams on this circuit.
     #[cfg(feature = "hs-service")]
     AwaitStreamRequest {
@@ -233,18 +241,27 @@ pub(crate) enum CtrlCmd {
         /// Oneshot channel to notify on completion.
         done: ReactorResultChannel<()>,
         /// The hop that is allowed to create streams.
-        hop_num: HopNum,
+        hop: TargetHop,
         /// A filter used to check requests before passing them on.
         #[educe(Debug(ignore))]
         #[cfg(feature = "hs-service")]
         filter: Box<dyn IncomingStreamRequestFilter>,
     },
+    /// Request the binding key of a target hop.
+    #[cfg(feature = "hs-service")]
+    GetBindingKey {
+        /// The hop for which we want the key.
+        hop: TargetHop,
+        /// Oneshot channel to notify on completion.
+        done: ReactorResultChannel<Option<CircuitBinding>>,
+    },
     /// (tests only) Add a hop to the list of hops on this circuit, with dummy cryptography.
     #[cfg(test)]
     AddFakeHop {
         relay_cell_format: RelayCellFormat,
         fwd_lasthop: bool,
         rev_lasthop: bool,
+        peer_id: path::HopDetail,
         params: CircParameters,
         done: ReactorResultChannel<()>,
     },
@@ -252,6 +269,7 @@ pub(crate) enum CtrlCmd {
     #[cfg(test)]
     QuerySendWindow {
         hop: HopNum,
+        leg: UniqId,
         done: ReactorResultChannel<(u32, Vec<SendmeTag>)>,
     },
     /// Shut down the reactor, and return the underlying [`Circuit`],
@@ -284,14 +302,19 @@ impl<'a> ControlHandler<'a> {
 
     /// Handle a control message.
     pub(super) fn handle_msg(&mut self, msg: CtrlMsg) -> Result<Option<RunOnceCmdInner>> {
-        trace!("{}: reactor received {:?}", self.reactor.unique_id, msg);
+        trace!(
+            tunnel_id = %self.reactor.tunnel_id,
+            msg = ?msg,
+            "reactor received control message"
+        );
+
         match msg {
             // This is handled earlier, since it requires blocking.
             CtrlMsg::Create { done, .. } => {
                 if self.reactor.circuits.len() == 1 {
                     // This should've been handled in Reactor::run_once()
                     // (ControlHandler::handle_msg() is never called before wait_for_create()).
-                    debug_assert!(self.reactor.circuits.single_leg()?.1.has_hops());
+                    debug_assert!(self.reactor.circuits.single_leg()?.has_hops());
                     // Don't care if the receiver goes away
                     let _ = done.send(Err(tor_error::bad_api_usage!(
                         "cannot create first hop twice"
@@ -311,10 +334,10 @@ impl<'a> ControlHandler<'a> {
                 peer_id,
                 public_key,
                 linkspecs,
-                params,
+                settings,
                 done,
             } => {
-                let Ok((leg_id, circ)) = self.reactor.circuits.single_leg_mut() else {
+                let Ok(circ) = self.reactor.circuits.single_leg_mut() else {
                     // Don't care if the receiver goes away
                     let _ = done.send(Err(tor_error::bad_api_usage!(
                         "cannot extend multipath tunnel"
@@ -330,7 +353,7 @@ impl<'a> ControlHandler<'a> {
                     HandshakeType::NTOR,
                     &public_key,
                     linkspecs,
-                    params,
+                    settings,
                     &(),
                     circ,
                     done,
@@ -340,7 +363,7 @@ impl<'a> ControlHandler<'a> {
                     .set_meta_handler(Box::new(extender))?;
 
                 Ok(Some(RunOnceCmdInner::Send {
-                    leg: leg_id,
+                    leg: circ.unique_id(),
                     cell,
                     done: None,
                 }))
@@ -349,10 +372,10 @@ impl<'a> ControlHandler<'a> {
                 peer_id,
                 public_key,
                 linkspecs,
-                params,
+                settings,
                 done,
             } => {
-                let Ok((leg_id, circ)) = self.reactor.circuits.single_leg_mut() else {
+                let Ok(circ) = self.reactor.circuits.single_leg_mut() else {
                     // Don't care if the receiver goes away
                     let _ = done.send(Err(tor_error::bad_api_usage!(
                         "cannot extend multipath tunnel"
@@ -365,7 +388,7 @@ impl<'a> ControlHandler<'a> {
                 // TODO #1067, TODO #1947: support negotiating other formats.
                 let relay_cell_format = RelayCellFormat::V0;
 
-                let client_extensions = circ_extensions_from_params(&params)?;
+                let client_extensions = circ_extensions_from_settings(&settings)?;
 
                 let (extender, cell) =
                     CircuitExtender::<NtorV3Client, Tor1RelayCrypto, _, _>::begin(
@@ -374,7 +397,7 @@ impl<'a> ControlHandler<'a> {
                         HandshakeType::NTOR_V3,
                         &public_key,
                         linkspecs,
-                        params,
+                        settings,
                         &client_extensions,
                         circ,
                         done,
@@ -384,7 +407,7 @@ impl<'a> ControlHandler<'a> {
                     .set_meta_handler(Box::new(extender))?;
 
                 Ok(Some(RunOnceCmdInner::Send {
-                    leg: leg_id,
+                    leg: circ.unique_id(),
                     cell,
                     done: None,
                 }))
@@ -449,20 +472,23 @@ impl<'a> ControlHandler<'a> {
                 done: Some(done),
             })),
             // TODO(#1860): remove stream-level sendme support
-            CtrlMsg::SendSendme {
-                stream_id,
-                hop,
-                sender,
-            } => {
-                // If resolving the hop fails,
-                // we want to report an error back to the initiator and not shut down the reactor.
+            CtrlMsg::SendSendme { stream_id, hop } => {
                 let (leg_id, hop_num) = match self.reactor.resolve_hop_location(hop) {
                     Ok(x) => x,
-                    Err(e) => {
-                        let e = into_bad_api_usage!("Could not resolve hop {hop:?}")(e);
-                        // don't care if receiver goes away
-                        let _ = sender.send(Err(e.into()));
-                        return Ok(None);
+                    Err(NoJoinPointError) => {
+                        // A stream tried to send a stream-level SENDME message to the join point of
+                        // a tunnel that has never had a join point. Currently in arti, only a
+                        // `StreamTarget` asks us to send a stream-level SENDME, and this tunnel
+                        // originally created the `StreamTarget` to begin with. So this is a
+                        // legitimate bug somewhere in the tunnel code.
+                        let err = internal!(
+                            "Could not send a stream-level SENDME to a join point on a tunnel without a join point",
+                        );
+                        // TODO: Rather than calling `warn_report` here, we should call
+                        // `trace_report!` from `Reactor::run_once()`. Since this is an internal
+                        // error, `trace_report!` should log it at "warn" level.
+                        warn_report!(err, "Tunnel reactor error");
+                        return Err(err.into());
                     }
                 };
 
@@ -470,18 +496,18 @@ impl<'a> ControlHandler<'a> {
                 let sendme_required = match self.reactor.uses_stream_sendme(leg_id, hop_num) {
                     Some(x) => x,
                     None => {
-                        // don't care if receiver goes away
-                        let _ = sender.send(Err(bad_api_usage!(
-                            "Unknown hop {hop_num:?} on leg {leg_id:?}"
-                        )
-                        .into()));
+                        // The leg/hop has disappeared. This is fine since the stream may have ended
+                        // and been cleaned up while this `CtrlMsg::SendSendme` message was queued.
+                        // It is possible that is a bug and this is an incorrect leg/hop number, but
+                        // it's not currently possible to differentiate between an incorrect leg/hop
+                        // number and a circuit hop that has been closed.
+                        debug!("Could not send a stream-level SENDME on a hop that does not exist. Ignoring.");
                         return Ok(None);
                     }
                 };
 
                 if !sendme_required {
-                    // don't care if receiver goes away
-                    let _ = sender.send(Ok(()));
+                    // Nothing to do, so discard the SENDME.
                     return Ok(None);
                 }
 
@@ -497,7 +523,7 @@ impl<'a> ControlHandler<'a> {
                 Ok(Some(RunOnceCmdInner::Send {
                     leg: leg_id,
                     cell,
-                    done: Some(sender),
+                    done: None,
                 }))
             }
             // TODO(conflux): this should specify which leg to send the msg on
@@ -506,11 +532,13 @@ impl<'a> ControlHandler<'a> {
             // This will involve updating ClientCIrc::send_raw_msg() to take a
             // leg id argument (which is a breaking change.
             #[cfg(feature = "send-control-msg")]
-            CtrlMsg::SendMsg {
-                hop_num,
-                msg,
-                sender,
-            } => {
+            CtrlMsg::SendMsg { hop, msg, sender } => {
+                let Some((leg_id, hop_num)) = self.reactor.target_hop_to_hopnum_id(hop) else {
+                    // Don't care if receiver goes away
+                    let _ = sender.send(Err(bad_api_usage!("Unknown {hop:?}").into()));
+                    return Ok(None);
+                };
+
                 let cell = AnyRelayMsgOuter::new(None, msg);
                 let cell = SendRelayCell {
                     hop: hop_num,
@@ -518,10 +546,8 @@ impl<'a> ControlHandler<'a> {
                     cell,
                 };
 
-                let leg = self.reactor.circuits.primary_leg_id();
-
                 Ok(Some(RunOnceCmdInner::Send {
-                    leg,
+                    leg: leg_id,
                     cell,
                     done: Some(sender),
                 }))
@@ -549,8 +575,14 @@ impl<'a> ControlHandler<'a> {
     }
 
     /// Handle a control command.
+    #[allow(clippy::needless_pass_by_value)] // Needed when conflux is enabled
     pub(super) fn handle_cmd(&mut self, msg: CtrlCmd) -> StdResult<(), ReactorError> {
-        trace!("{}: reactor received {:?}", self.reactor.unique_id, msg);
+        trace!(
+            tunnel_id = %self.reactor.tunnel_id,
+            msg = ?msg,
+            "reactor received control command"
+        );
+
         match msg {
             CtrlCmd::Shutdown => self.reactor.handle_shutdown().map(|_| ()),
             #[cfg(feature = "hs-common")]
@@ -558,7 +590,7 @@ impl<'a> ControlHandler<'a> {
             CtrlCmd::ExtendVirtual {
                 relay_cell_format: format,
                 cell_crypto,
-                params,
+                settings,
                 done,
             } => {
                 let (outbound, inbound, binding) = cell_crypto;
@@ -567,7 +599,7 @@ impl<'a> ControlHandler<'a> {
                 // describe why the virtual hop was added, or something?
                 let peer_id = path::HopDetail::Virtual;
 
-                let Ok((_id, leg)) = self.reactor.circuits.single_leg_mut() else {
+                let Ok(leg) = self.reactor.circuits.single_leg_mut() else {
                     // Don't care if the receiver goes away
                     let _ = done.send(Err(tor_error::bad_api_usage!(
                         "cannot extend multipath tunnel"
@@ -577,19 +609,31 @@ impl<'a> ControlHandler<'a> {
                     return Ok(());
                 };
 
-                leg.add_hop(format, peer_id, outbound, inbound, binding, &params)?;
+                leg.add_hop(format, peer_id, outbound, inbound, binding, &settings)?;
                 let _ = done.send(Ok(()));
 
                 Ok(())
             }
+            CtrlCmd::ResolveTargetHop { hop, done } => {
+                let _ = done.send(
+                    self.reactor
+                        .resolve_target_hop(hop)
+                        .map_err(|_| crate::util::err::Error::NoSuchHop),
+                );
+                Ok(())
+            }
             #[cfg(feature = "hs-service")]
             CtrlCmd::AwaitStreamRequest {
                 cmd_checker,
                 incoming_sender,
-                hop_num,
+                hop,
                 done,
                 filter,
             } => {
+                let Some((_, hop_num)) = self.reactor.target_hop_to_hopnum_id(hop) else {
+                    let _ = done.send(Err(crate::Error::NoSuchHop));
+                    return Ok(());
+                };
                 // TODO: At some point we might want to add a CtrlCmd for
                 // de-registering the handler.  See comments on `allow_stream_requests`.
                 let handler = IncomingStreamRequestHandler {
@@ -607,15 +651,38 @@ impl<'a> ControlHandler<'a> {
 
                 Ok(())
             }
+            #[cfg(feature = "hs-service")]
+            CtrlCmd::GetBindingKey { hop, done } => {
+                let Some((leg_id, hop_num)) = self.reactor.target_hop_to_hopnum_id(hop) else {
+                    let _ = done.send(Err(tor_error::internal!(
+                        "Unknown TargetHop when getting binding key"
+                    )
+                    .into()));
+                    return Ok(());
+                };
+                let Some(circuit) = self.reactor.circuits.leg(leg_id) else {
+                    let _ = done.send(Err(tor_error::bad_api_usage!(
+                        "Unknown circuit id {leg_id} when getting binding key"
+                    )
+                    .into()));
+                    return Ok(());
+                };
+                // Get the binding key from the mutable state and send it back.
+                let key = circuit.mutable().binding_key(hop_num);
+                let _ = done.send(Ok(key));
+
+                Ok(())
+            }
             #[cfg(test)]
             CtrlCmd::AddFakeHop {
                 relay_cell_format,
                 fwd_lasthop,
                 rev_lasthop,
+                peer_id,
                 params,
                 done,
             } => {
-                let Ok((_id, leg)) = self.reactor.circuits.single_leg_mut() else {
+                let Ok(leg) = self.reactor.circuits.single_leg_mut() else {
                     // Don't care if the receiver goes away
                     let _ = done.send(Err(tor_error::bad_api_usage!(
                         "cannot add fake hop to multipath tunnel"
@@ -625,21 +692,24 @@ impl<'a> ControlHandler<'a> {
                     return Ok(());
                 };
 
-                leg.handle_add_fake_hop(relay_cell_format, fwd_lasthop, rev_lasthop, &params, done);
+                leg.handle_add_fake_hop(
+                    relay_cell_format,
+                    fwd_lasthop,
+                    rev_lasthop,
+                    peer_id,
+                    &params,
+                    done,
+                );
 
                 Ok(())
             }
             #[cfg(test)]
-            CtrlCmd::QuerySendWindow { hop, done } => {
+            CtrlCmd::QuerySendWindow { hop, leg, done } => {
                 // Immediately invoked function means that errors will be sent to the channel.
                 let _ = done.send((|| {
-                    let (_id, leg) =
-                        self.reactor
-                            .circuits
-                            .single_leg_mut()
-                            .map_err(into_bad_api_usage!(
-                                "cannot query send window of multipath tunnel"
-                            ))?;
+                    let leg = self.reactor.circuits.leg_mut(leg).ok_or_else(|| {
+                        bad_api_usage!("cannot query send window of non-existent circuit")
+                    })?;
 
                     let hop = leg.hop_mut(hop).ok_or(bad_api_usage!(
                         "received QuerySendWindow for unknown hop {}",
