@@ -2,7 +2,7 @@
 use std::{sync::Arc, time::Duration};
 use thiserror::Error;
 use tor_cell::relaycell::{msg::EndReason, StreamId};
-use tor_error::{ErrorKind, HasKind};
+use tor_error::{Bug, ErrorKind, HasKind};
 use tor_linkspec::RelayIdType;
 
 /// An error type for the tor-proto crate.
@@ -306,16 +306,25 @@ pub(crate) enum ReactorError {
     /// The reactor should shut down without an error, since all is well.
     Shutdown,
 }
+
 impl From<Error> for ReactorError {
     fn from(e: Error) -> ReactorError {
         ReactorError::Err(e)
     }
 }
+
 impl From<ChannelClosed> for ReactorError {
     fn from(e: ChannelClosed) -> ReactorError {
         ReactorError::Err(e.into())
     }
 }
+
+impl From<Bug> for ReactorError {
+    fn from(e: Bug) -> ReactorError {
+        ReactorError::Err(e.into())
+    }
+}
+
 #[cfg(test)]
 impl ReactorError {
     /// Tests only: assert that this is an Error, and return it.
@@ -326,3 +335,16 @@ impl ReactorError {
         }
     }
 }
+
+/// An error type for client-side conflux handshakes.
+#[derive(Debug)]
+#[cfg(feature = "conflux")]
+#[allow(unused)] // TODO(conflux): remove
+pub(crate) enum ConfluxHandshakeError {
+    /// Timeout while waiting for CONFLUX_LINKED response.
+    Timeout,
+    /// An error that occurred while sending the CONFLUX_LINK message.
+    Link(Error),
+    /// The channel was closed.
+    ChannelClosed,
+}
