@@ -10,6 +10,7 @@
 //! See [`MockSleepProvider`] for more information.
 //! Use [`MockRuntime`](crate::MockRuntime) for new tests.
 
+#![forbid(unsafe_code)] // if you remove this, enable (or write) miri tests (git grep miri)
 #![allow(clippy::missing_docs_in_private_items)]
 
 use std::{
@@ -69,11 +70,6 @@ use crate::time_core::MockTimeCore;
 ///    coordination with the executor is required.
 ///    This coordination is provided by the integrated `MockRuntime`;
 ///    `SimpleMockTimeProvider` is of limited usefulness by itself.
-//
-// TODO: at some point we should add #[deprecated] to this type
-// and to the block_advance etc. methods in SleepProvider.
-// But right now that would involve rewriting a whole bunch of tests,
-// or generous sprinklings of #[allow].
 ///
 /// ### Examples
 ///
@@ -151,6 +147,10 @@ use crate::time_core::MockTimeCore;
 /// });
 /// ```
 #[derive(Clone)]
+// When we're used by external crates, we're always cfg(not(test)), so we seem deprecated
+// from outside this crate.  *Within* this crate, this cfg_attr means that if we use things
+// that are deprecated for other reasons, we will notice.
+#[cfg_attr(not(test), deprecated(since = "0.29.0"))]
 pub struct MockSleepProvider {
     /// The shared backend for this MockSleepProvider and its futures.
     state: Arc<Mutex<SleepSchedule>>,
@@ -577,7 +577,7 @@ impl Future for Sleeping {
     }
 }
 
-#[cfg(test)]
+#[cfg(all(test, not(miri)))] // miri cannot do CLOCK_REALTIME
 mod test {
     // @@ begin test lint list maintained by maint/add_warning @@
     #![allow(clippy::bool_assert_comparison)]
